@@ -1,4 +1,5 @@
 import ctypes
+import locale
 import logging
 import os
 import sys
@@ -10,8 +11,12 @@ logger = logging.getLogger(__name__)
 
 TESSDATA_PREFIX = os.getenv('TESSDATA_PREFIX', None)
 libnames = []
+# 70 is the minimum credible dpi for tesseract and force it to compute an
+# estimate of the image dpi
+DPI_DEFAULT = 70
 
-if getattr(sys, 'frozen', False):
+
+if getattr(sys, 'frozen', False):  # pragma: no cover
     # Pyinstaller integration
     libnames += [os.path.join(sys._MEIPASS, "libtesseract-4.dll")]
     libnames += [os.path.join(sys._MEIPASS, "libtesseract-3.dll")]
@@ -23,15 +28,21 @@ if getattr(sys, 'frozen', False):
             )
         )
     else:
-        TESSDATA_PREFIX = tessdata
+        TESSDATA_PREFIX = os.path.join(tessdata, "tessdata")
 
 
-if sys.platform[:3] == "win":
+if sys.platform[:3] == "win":  # pragma: no cover
     libnames += [
         # Jflesch> Don't they have the equivalent of LD_LIBRARY_PATH on
         # Windows ?
         "../vs2010/DLL_Release/libtesseract302.dll",
+        # prefer the most recent first
+        "libtesseract305.dll",
+        "libtesseract304.dll",
+        "libtesseract303.dll",
         "libtesseract302.dll",
+        "libtesseract400.dll",  # Tesseract 4 is still in alpha stage
+        "libtesseract.dll",
         "C:\\Program Files (x86)\\Tesseract-OCR\\libtesseract-4.dll",
         "C:\\Program Files (x86)\\Tesseract-OCR\\libtesseract-3.dll",
     ]
@@ -44,12 +55,19 @@ else:
 
 g_libtesseract = None
 
-for libname in libnames:
+lib_load_errors = []
+for libname in libnames:  # pragma: no branch
     try:
         g_libtesseract = ctypes.cdll.LoadLibrary(libname)
+        lib_load_errors = []
         break
-    except OSError:
-        pass
+    except OSError as ex:  # pragma: no cover
+        if hasattr(ex, 'message'):
+            # python 2
+            lib_load_errors.append((libname, ex.message))
+        else:
+            # python 3
+            lib_load_errors.append((libname, str(ex)))
 
 
 class PageSegMode(object):
@@ -66,7 +84,8 @@ class PageSegMode(object):
     SINGLE_CHAR = 10
     SPARSE_TEXT = 11
     SPARSE_TEXT_OSD = 12
-    COUNT = 13
+    PSM_RAW_LINE = 13
+    COUNT = 14
 
 
 class Orientation(object):
@@ -115,7 +134,7 @@ class OSResults(ctypes.Structure):
     ]
 
 
-if g_libtesseract:
+if g_libtesseract:  # pragma: no cover
     g_libtesseract.TessVersion.argtypes = []
     g_libtesseract.TessVersion.restype = ctypes.c_char_p
 
@@ -126,6 +145,12 @@ if g_libtesseract:
     ]
     g_libtesseract.TessBaseAPIDelete.argtypes = None
 
+    g_libtesseract.TessBaseAPIGetDatapath.argtypes = [
+        ctypes.c_void_p,  # TessBaseAPI*
+    ]
+    g_libtesseract.TessBaseAPIGetDatapath.restype = ctypes.POINTER(
+        ctypes.c_char)
+
     g_libtesseract.TessBaseAPIInit1.argtypes = [
         ctypes.c_void_p,  # TessBaseAPI*
         ctypes.c_char_p,  # datapath
@@ -134,7 +159,7 @@ if g_libtesseract:
         ctypes.POINTER(ctypes.c_char_p),  # configs
         ctypes.c_int,  # configs_size
     ]
-    g_libtesseract.TessBaseAPIInit1.restypes = ctypes.c_int
+    g_libtesseract.TessBaseAPIInit1.restype = ctypes.c_int
 
     g_libtesseract.TessBaseAPIInit3.argtypes = [
         ctypes.c_void_p,  # TessBaseAPI*
@@ -143,6 +168,13 @@ if g_libtesseract:
     ]
     g_libtesseract.TessBaseAPIInit3.restype = ctypes.c_int
 
+    g_libtesseract.TessBaseAPISetSourceResolution.argtypes = [
+        ctypes.c_void_p,  # TessBaseAPI*
+        ctypes.c_int,     # PPI
+    ]
+
+    g_libtesseract.TessBaseAPISetSourceResolution.restype = None
+
     g_libtesseract.TessBaseAPISetVariable.argtypes = [
         ctypes.c_void_p,  # TessBaseAPI*
         ctypes.c_char_p,  # name
@@ -177,6 +209,36 @@ if g_libtesseract:
     ]
     g_libtesseract.TessBaseAPISetImage.restype = None
 
+    g_libtesseract.TessResultRendererAddImage.argtypes = [
+        ctypes.c_void_p,  # TessResultRenderer* renderer
+        ctypes.c_void_p  # TessBaseAPI* api
+    ]
+    g_libtesseract.TessResultRendererAddImage.restype = ctypes.c_bool
+
+    g_libtesseract.TessBaseAPISetInputName.argtypes = [
+        ctypes.c_void_p,  # TessBaseAPI* handle
+        ctypes.c_char_p  # const char* name
+    ]
+    g_libtesseract.TessBaseAPISetInputName.restype = None
+
+    g_libtesseract.TessResultRendererBeginDocument.argtypes = [
+        ctypes.c_void_p,  # TessResultRenderer* renderer
+        ctypes.c_char_p  # const char* title
+    ]
+    g_libtesseract.TessResultRendererBeginDocument.restype = ctypes.c_bool
+
+    g_libtesseract.TessResultRendererEndDocument.argtypes = [
+        ctypes.c_void_p  # TessResultRenderer* renderer
+    ]
+    g_libtesseract.TessResultRendererEndDocument.restype = ctypes.c_bool
+
+    g_libtesseract.TessPDFRendererCreate.argtypes = [
+        ctypes.c_char_p,  # const char* outputbase
+        ctypes.c_char_p,  # const char* datadir
+        ctypes.c_bool  # BOOL textonly
+    ]
+    g_libtesseract.TessPDFRendererCreate.restype = ctypes.c_void_p
+
     g_libtesseract.TessBaseAPIRecognize.argtypes = [
         ctypes.c_void_p,  # TessBaseAPI*
         ctypes.c_void_p,  # ETEXT_DESC*
@@ -262,6 +324,12 @@ if g_libtesseract:
     g_libtesseract.TessResultIteratorGetUTF8Text.restype = \
         ctypes.c_void_p
 
+    g_libtesseract.TessResultIteratorConfidence.argtypes = [
+        ctypes.c_void_p,
+        ctypes.c_int,
+    ]
+    g_libtesseract.TessResultIteratorConfidence.restype = ctypes.c_float
+
     g_libtesseract.TessDeleteText.argtypes = [
         ctypes.c_void_p
     ]
@@ -287,12 +355,17 @@ if g_libtesseract:
 
 def init(lang=None):
     assert(g_libtesseract)
+
+    # Tesseract 4 workaround
+    if get_version() == "4.0.0":
+        locale.setlocale(locale.LC_ALL, "C")
+
     handle = g_libtesseract.TessBaseAPICreate()
     try:
         if lang:
             lang = lang.encode("utf-8")
         prefix = None
-        if TESSDATA_PREFIX:
+        if TESSDATA_PREFIX:  # pragma: no cover
             prefix = TESSDATA_PREFIX.encode("utf-8")
         g_libtesseract.TessBaseAPIInit3(
             ctypes.c_void_p(handle),
@@ -304,30 +377,27 @@ def init(lang=None):
             b"tessedit_zero_rejection",
             b"F"
         )
-    except:
+    except:  # noqa: E722
         g_libtesseract.TessBaseAPIDelete(ctypes.c_void_p(handle))
         raise
     return handle
 
 
 def cleanup(handle):
+    assert(g_libtesseract)
     g_libtesseract.TessBaseAPIDelete(ctypes.c_void_p(handle))
 
 
 def is_available():
-    global g_libtesseract
     return g_libtesseract is not None
 
 
 def get_version():
-    global g_libtesseract
     assert(g_libtesseract)
-
     return g_libtesseract.TessVersion().decode("utf-8")
 
 
 def get_available_languages(handle):
-    global g_libtesseract
     assert(g_libtesseract)
 
     langs = []
@@ -343,7 +413,6 @@ def get_available_languages(handle):
 
 
 def set_is_numeric(handle, mode):
-    global g_libtesseract
     assert(g_libtesseract)
 
     if mode:
@@ -358,8 +427,20 @@ def set_is_numeric(handle, mode):
     )
 
 
+def set_debug_file(handle, filename):
+    assert(g_libtesseract)
+
+    if not isinstance(filename, bytes):
+        filename = filename.encode('utf-8')
+
+    g_libtesseract.TessBaseAPISetVariable(
+        ctypes.c_void_p(handle),
+        b"debug_file",
+        filename
+    )
+
+
 def set_page_seg_mode(handle, mode):
-    global g_libtesseract
     assert(g_libtesseract)
 
     g_libtesseract.TessBaseAPISetPageSegMode(
@@ -368,34 +449,32 @@ def set_page_seg_mode(handle, mode):
 
 
 def init_for_analyse_page(handle):
-    global g_libtesseract
     assert(g_libtesseract)
 
     g_libtesseract.TessBaseAPIInitForAnalysePage(ctypes.c_void_p(handle))
 
 
 def set_image(handle, image):
-    global g_libtesseract
     assert(g_libtesseract)
 
     image = image.convert("RGB")
     image.load()
     imgdata = image.tobytes("raw", "RGB")
 
-    imgsize = image.size
-
     g_libtesseract.TessBaseAPISetImage(
         ctypes.c_void_p(handle),
         imgdata,
-        ctypes.c_int(imgsize[0]),
-        ctypes.c_int(imgsize[1]),
+        ctypes.c_int(image.width),
+        ctypes.c_int(image.height),
         ctypes.c_int(3),  # RGB = 3 * 8
-        ctypes.c_int(imgsize[0] * 3)
+        ctypes.c_int(image.width * 3)
     )
 
+    dpi = image.info.get("dpi", [DPI_DEFAULT])[0]
+    g_libtesseract.TessBaseAPISetSourceResolution(ctypes.c_void_p(handle), dpi)
+
 
 def recognize(handle):
-    global g_libtesseract
     assert(g_libtesseract)
 
     return g_libtesseract.TessBaseAPIRecognize(
@@ -404,13 +483,13 @@ def recognize(handle):
 
 
 def analyse_layout(handle):
-    global g_libtesseract
     assert(g_libtesseract)
 
     return g_libtesseract.TessBaseAPIAnalyseLayout(ctypes.c_void_p(handle))
 
 
 def get_utf8_text(handle):
+    assert(g_libtesseract)
     ptr = g_libtesseract.TessBaseAPIGetUTF8Text(ctypes.c_void_p(handle))
     val = ctypes.cast(ptr, ctypes.c_char_p).value.decode("utf-8")
     g_libtesseract.TessDeleteText(ptr)
@@ -418,21 +497,19 @@ def get_utf8_text(handle):
 
 
 def page_iterator_delete(iterator):
-    global g_libtesseract
     assert(g_libtesseract)
 
     return g_libtesseract.TessPageIteratorDelete(ctypes.c_void_p(iterator))
 
 
 def page_iterator_next(iterator, level):
-    global g_libtesseract
     assert(g_libtesseract)
 
-    return g_libtesseract.TessPageIteratorNext(ctypes.c_void_p(iterator), level)
+    return g_libtesseract.TessPageIteratorNext(ctypes.c_void_p(iterator),
+                                               level)
 
 
 def page_iterator_is_at_beginning_of(iterator, level):
-    global g_libtesseract
     assert(g_libtesseract)
 
     return g_libtesseract.TessPageIteratorIsAtBeginningOf(
@@ -441,7 +518,6 @@ def page_iterator_is_at_beginning_of(iterator, level):
 
 
 def page_iterator_is_at_final_element(iterator, level, element):
-    global g_libtesseract
     assert(g_libtesseract)
 
     return g_libtesseract.TessPageIteratorIsAtFinalElement(
@@ -450,7 +526,6 @@ def page_iterator_is_at_final_element(iterator, level, element):
 
 
 def page_iterator_block_type(iterator):
-    global g_libtesseract
     assert(g_libtesseract)
 
     return g_libtesseract.TessPageIteratorBlockType(
@@ -459,7 +534,6 @@ def page_iterator_block_type(iterator):
 
 
 def page_iterator_bounding_box(iterator, level):
-    global g_libtesseract
     assert(g_libtesseract)
 
     left = ctypes.c_int(0)
@@ -485,7 +559,6 @@ def page_iterator_bounding_box(iterator, level):
 
 
 def page_iterator_orientation(iterator):
-    global g_libtesseract
     assert(g_libtesseract)
 
     orientation = ctypes.c_int(0)
@@ -510,7 +583,6 @@ def page_iterator_orientation(iterator):
 
 
 def get_iterator(handle):
-    global g_libtesseract
     assert(g_libtesseract)
 
     i = g_libtesseract.TessBaseAPIGetIterator(ctypes.c_void_p(handle))
@@ -518,7 +590,6 @@ def get_iterator(handle):
 
 
 def result_iterator_get_page_iterator(res_iterator):
-    global g_libtesseract
     assert(g_libtesseract)
 
     return g_libtesseract.TessResultIteratorGetPageIterator(
@@ -527,6 +598,7 @@ def result_iterator_get_page_iterator(res_iterator):
 
 
 def result_iterator_get_utf8_text(iterator, level):
+    assert(g_libtesseract)
     ptr = g_libtesseract.TessResultIteratorGetUTF8Text(
         ctypes.c_void_p(iterator), level
     )
@@ -537,8 +609,18 @@ def result_iterator_get_utf8_text(iterator, level):
     return val
 
 
+def result_iterator_get_confidence(iterator, level):
+    assert(g_libtesseract)
+    ptr = g_libtesseract.TessResultIteratorConfidence(
+        ctypes.c_void_p(iterator), level
+    )
+    if ptr is None:
+        return None
+    val = ctypes.c_float(ptr).value
+    return val
+
+
 def detect_os(handle):
-    global g_libtesseract
     assert(g_libtesseract)
 
     # Use the new API function if it is available, because since Tesseract
@@ -575,3 +657,52 @@ def detect_os(handle):
             "orientation": results.best_orientation_id,
             "confidence": results.best_oconfidence,
         }
+
+
+def set_input_name(handle, input_file):
+    assert(g_libtesseract)
+
+    g_libtesseract.TessBaseAPISetInputName(
+        ctypes.c_void_p(handle),
+        input_file.encode()
+    )
+
+
+def init_pdf_renderer(handle, output_file, textonly):
+    assert(g_libtesseract)
+
+    tessdata_dir = g_libtesseract.TessBaseAPIGetDatapath(handle)
+
+    renderer = g_libtesseract.TessPDFRendererCreate(
+        output_file.encode(),
+        tessdata_dir,
+        ctypes.c_bool(textonly)
+    )
+
+    return renderer
+
+
+def begin_document(renderer, doc_name):
+    assert(g_libtesseract)
+
+    g_libtesseract.TessResultRendererBeginDocument(
+        ctypes.c_void_p(renderer),
+        doc_name.encode()
+    )
+
+
+def add_renderer_image(handle, renderer):
+    assert(g_libtesseract)
+
+    g_libtesseract.TessResultRendererAddImage(
+        ctypes.c_void_p(renderer),
+        ctypes.c_void_p(handle)
+    )
+
+
+def end_document(renderer):
+    assert(g_libtesseract)
+
+    g_libtesseract.TessResultRendererEndDocument(
+        ctypes.c_void_p(renderer)
+    )
